@@ -18,8 +18,6 @@ import weakref
 from xml.etree.ElementTree import fromstring as rfromstring
 import zipfile
 
-import six
-
 import pyghmi.constants as pygconst
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.private.session as ipmisession
@@ -527,8 +525,11 @@ class SMMClient(object):
             }
             for baynum in range(numbays):
                 baynum += 1
-                baycapinfo = self.ipmicmd.xraw_command(
-                    0x32, 0x9d, data=[baynum])
+                try:
+                    baycapinfo = self.ipmicmd.xraw_command(
+                        0x32, 0x9d, data=[baynum])
+                except Exception:
+                    continue
                 capmin, capmax, protcap, usercap, thermcap = struct.unpack(
                     '<HHHHH', baycapinfo['data'][retoffset:retoffset + 10])
                 settings['bay{0}_user_cap'.format(baynum)] = {
@@ -543,8 +544,11 @@ class SMMClient(object):
                     'help': 'Show the current protective cap for the system '
                             'in bay {0}'.format(baynum)
                 }
-                baycapstate = self.ipmicmd.xraw_command(
-                    0x32, 0xa0, data=[baynum])
+                try:
+                    baycapstate = self.ipmicmd.xraw_command(
+                        0x32, 0xa0, data=[baynum])
+                except Exception:
+                    continue
                 baycapstate = bool(baycapstate['data'][retoffset])
                 settings['bay{0}_user_cap_active'.format(baynum)] = {
                     'value': 'Enable' if baycapstate else 'Disable',
@@ -635,7 +639,7 @@ class SMMClient(object):
         for key in changeset:
             if not key:
                 raise pygexc.InvalidParameterValue('Empty key is invalid')
-            if isinstance(changeset[key], six.string_types):
+            if isinstance(changeset[key], str):
                 changeset[key] = {'value': changeset[key]}
             for rule in self.rulemap:
                 if fnmatch.fnmatch(rule, key.lower()):
@@ -756,7 +760,13 @@ class SMMClient(object):
                 rsp = {'data': [1]}
             rsp['data'] = bytearray(rsp['data'])
             if rsp['data'][0] == 2:  # shared io
-                rsp = self.ipmicmd.xraw_command(0x32, 0xa7, data=[bay - 1])
+                try:
+                    rsp = self.ipmicmd.xraw_command(0x32, 0xa7, data=[bay - 1])
+                except Exception:
+                    raise Exception('Shared IO detected trying to reseat {}, '
+                                    'but unable to determine status of '
+                                    'partner bay {}'.format(
+                                        bay, bay - 1))
                 rsp['data'] = bytearray(rsp['data'])
                 if rsp['data'][1] == 0x80:
                     raise Exception('Unable to reseat bay {0} due to bay {1} '
