@@ -23,7 +23,6 @@ import re
 import socket
 import time
 
-import six
 import zipfile
 
 import pyghmi.constants as pygconst
@@ -119,6 +118,7 @@ def str_to_size(sizestr):
 
 
 class OEMHandler(generic.OEMHandler):
+    usegenericsensors = False
     logouturl = '/api/providers/logout'
     bmcname = 'XCC'
     ADP_URL = '/api/dataset/imm_adapters?params=pci_GetAdapters'
@@ -149,6 +149,31 @@ class OEMHandler(generic.OEMHandler):
         self.fwc = None
         self.fwo = None
 
+    def get_screenshot(self, outfile):
+        self.wc.grab_json_response('/api/providers/rp_screenshot')
+        url = '/download/HostScreenShot.png'
+        fd = webclient.FileDownloader(self.wc, url, outfile)
+        fd.start()
+        fd.join()
+
+    def get_extended_bmc_configuration(self, fishclient, hideadvanced=True):
+        immsettings = self.get_system_configuration(fetchimm=True, hideadvanced=hideadvanced)
+        for setting in list(immsettings):
+            if not setting.startswith('IMM.'):
+                del immsettings[setting]
+        return immsettings
+
+    def get_ikvm_methods(self):
+        return ['url']
+
+    def get_ikvm_launchdata(self):
+        access = self._do_web_request('/redfish/v1/Managers/1/Oem/Lenovo/RemoteControl/Actions/LenovoRemoteControlService.GetRemoteConsoleToken', {})
+        if access.get('Token', None):
+            accessinfo = {
+                'url': '/#/login?{}&context=remote&mode=multi'.format(access['Token'])
+                }
+            return accessinfo
+
     def get_system_configuration(self, hideadvanced=True, fishclient=None,
                                  fetchimm=False):
         if not self.fwc:
@@ -235,7 +260,7 @@ class OEMHandler(generic.OEMHandler):
 
     def merge_changeset(self, changeset):
         for key in changeset:
-            if isinstance(changeset[key], six.string_types):
+            if isinstance(changeset[key], str):
                 changeset[key] = {'value': changeset[key]}
             newvalue = changeset[key]['value']
             if self.fwo[key]['is_list'] and not isinstance(newvalue, list):
@@ -285,8 +310,12 @@ class OEMHandler(generic.OEMHandler):
             raise pygexc.UnsupportedFunctionality(
                 'This is not an enclosure manager')
         wc = self.wc.dupe(timeout=5)
-        rsp = wc.grab_json_response_with_status(
-            '/api/providers/virt_reseat', '{}')
+        try:
+            rsp = wc.grab_json_response_with_status(
+                '/api/providers/virt_reseat', '{}')
+        except socket.timeout:
+            # Can't be certain, but most likely a timeout'
+            return
         if rsp[1] == 500 and rsp[0] == 'Target Unavailable':
             return
         if rsp[1] != 200 or rsp[0].get('return', 1) != 0:
@@ -386,7 +415,7 @@ class OEMHandler(generic.OEMHandler):
         usbsettings = {}
         secparms = {}
         for key in changeset:
-            if isinstance(changeset[key], six.string_types):
+            if isinstance(changeset[key], str):
                 changeset[key] = {'value': changeset[key]}
             currval = changeset[key].get('value', None)
             if key.lower() in self.rulemap:
@@ -572,7 +601,7 @@ class OEMHandler(generic.OEMHandler):
         summary['badreadings'] = fallbackdata
         return summary
 
-    def get_description(self):
+    def get_description(self, fishclient):
         description = self._do_web_request('/DeviceDescription.json')
         if description:
             description = description[0]
@@ -630,7 +659,9 @@ class OEMHandler(generic.OEMHandler):
                     skipkeys.add(fwi['key'])
                     if fwi.get('fw_status', 0) == 2:
                         bdata = {}
-                        if 'fw_version_pend' in fwi:
+                        if 'fw_pkg_version' in fwi and fwi['fw_pkg_version']:
+                            bdata['version'] = fwi['fw_pkg_version']
+                        elif 'fw_version_pend' in fwi:
                             bdata['version'] = fwi['fw_version_pend']
                         yield '{0} Pending Update'.format(aname), bdata
         for fwi in fdata.get('items', []):
@@ -667,33 +698,57 @@ class OEMHandler(generic.OEMHandler):
             for diskent in adp.get('aimDisks', ()):
                 yield self._get_disk_firmware_single(diskent)
 
-    def get_firmware_inventory(self, components):
-        sysinf = self.wc.grab_json_response('/api/dataset/sys_info')
-        for item in sysinf.get('items', {}):
-            for firm in item.get('firmware', []):
-                firminfo = {
-                    'version': firm['version'],
-                    'build': firm['build'],
-                    'date': parse_time(firm['release_date']),
-                }
-                if firm['type'] == 5:
-                    yield ('XCC', firminfo)
-                elif firm['type'] == 6:
-                    yield ('XCC Backup', firminfo)
-                elif firm['type'] == 0:
-                    yield ('UEFI', firminfo)
-                elif firm['type'] == 7:
-                    yield ('LXPM', firminfo)
-                elif firm['type'] == 8:
-                    yield ('LXPM Windows Driver Bundle', firminfo)
-                elif firm['type'] == 9:
-                    yield ('LXPM Linux Driver Bundle', firminfo)
-                elif firm['type'] == 10:
-                    yield ('LXUM', firminfo)
-        for adpinfo in self._get_agentless_firmware(components):
-            yield adpinfo
-        for adpinfo in self._get_disk_firmware(components):
-            yield adpinfo
+    def get_firmware_inventory(self, components, fishclient, category=None):
+
+        if components and 'core' in components:
+            category = 'core'
+            components = components - set(['core'])
+        if not category:
+            category = 'all'
+        if category in ('all', 'core'):
+            sysinf = self.wc.grab_json_response('/api/dataset/sys_info')
+            for item in sysinf.get('items', {}):
+                for firm in item.get('firmware', []):
+                    firminfo = {
+                        'version': firm['version'],
+                        'build': firm['build'],
+                        'date': parse_time(firm['release_date']),
+                    }
+                    if firm['type'] == 5:
+                        yield ('XCC', firminfo)
+                    elif firm['type'] == 6:
+                        yield ('XCC Backup', firminfo)
+                    elif firm['type'] == 0:
+                        yield ('UEFI', firminfo)
+                    elif firm['type'] == 7:
+                        yield ('LXPM', firminfo)
+                    elif firm['type'] == 8:
+                        yield ('LXPM Windows Driver Bundle', firminfo)
+                    elif firm['type'] == 9:
+                        yield ('LXPM Linux Driver Bundle', firminfo)
+                    elif firm['type'] == 10:
+                        yield ('LXUM', firminfo)
+        if components:
+            components = set([x.lower() for x in components])
+            components = components - set((
+                'core', 'uefi', 'bios', 'xcc', 'bmc', 'imm', 'fpga',
+                'lxpm'))
+            if not components:
+                return
+        if not components:
+            components = set(('all',))
+        needadapterfirmware = False
+        needdiskfirmware = False
+        if category in ('all', 'adapters'):
+            needadapterfirmware = True
+        if category in ('all', 'disks'):
+            needdiskfirmware = True
+        if needadapterfirmware:
+            for adpinfo in self._get_agentless_firmware(components):
+                yield adpinfo
+        if needdiskfirmware:
+            for adpinfo in self._get_disk_firmware(components):
+                yield adpinfo
         raise pygexc.BypassGenericBehavior()
 
     def get_storage_configuration(self, logout=True):
@@ -1119,7 +1174,7 @@ class OEMHandler(generic.OEMHandler):
         res = wc.grab_json_response_with_status(url, body, method=method)
         return res
 
-    def list_media(self, fishclient):
+    def list_media(self, fishclient, cache=True):
         rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_getdisk')
         if 'items' in rt:
             for mt in rt['items']:
@@ -1131,6 +1186,10 @@ class OEMHandler(generic.OEMHandler):
                     url = url.replace(':', '')
                     url = 'nfs://' + url
                 yield media.Media(mt['filename'], url)
+        for rdoc in self._list_rdoc():
+            yield rdoc
+    
+    def _list_rdoc(self):
         rt = self.wc.grab_json_response('/api/providers/rp_rdoc_imagelist')
         if 'items' in rt:
             for mt in rt['items']:
@@ -1154,6 +1213,15 @@ class OEMHandler(generic.OEMHandler):
     def upload_media(self, filename, progress=None, data=None):
         wc = self.wc
         self._refresh_token()
+        numrdocs = 0
+        for rdoc in self._list_rdoc():
+            numrdocs += 1
+            if rdoc.name == os.path.basename(filename):
+                raise pygexc.InvalidParameterValue(
+                    'An image with that name already exists')
+        if numrdocs >= 2:
+            raise pygexc.InvalidParameterValue(
+                'Maximum number of uploaded media reached')
         rsp, statu = wc.grab_json_response_with_status('/rdocupload')
         newmode = False
         if statu == 404:
@@ -1178,10 +1246,16 @@ class OEMHandler(generic.OEMHandler):
                     progress({'phase': 'upload',
                             'progress': 100.0 * rsp['received'] / rsp['size']})
             self._refresh_token()
-        rsp = json.loads(uploadthread.rsp)
+        if uploadthread.rsp:
+            rsp = json.loads(uploadthread.rsp)
+        else:
+            rsp = {}
         if progress:
             progress({'phase': 'upload',
                       'progress': 100.0})
+        if 'items' not in rsp or len(rsp['items']) == 0:
+            errmsg = repr(rsp) if rsp else self.wc.lastjsonerror if self.wc.lastjsonerror else repr(uploadthread.rspstatus)
+            raise pygexc.PyghmiException('Failed to upload image: ' + errmsg)
         thepath = rsp['items'][0]['path']
         thename = rsp['items'][0]['name']
         writeable = 1 if filename.lower().endswith('.img') else 0
@@ -1193,7 +1267,7 @@ class OEMHandler(generic.OEMHandler):
         self._refresh_token()
         if rsp.get('return', -1) != 0:
             errmsg = repr(rsp) if rsp else self.wc.lastjsonerror
-            raise Exception('Unrecognized return: ' + errmsg)
+            raise pygexc.PyghmiException('Failed to upload image: ' + errmsg)
         ready = False
         while not ready:
             time.sleep(3)
@@ -1264,24 +1338,34 @@ class OEMHandler(generic.OEMHandler):
             if (uploadthread.rspstatus >= 300
                     or uploadthread.rspstatus < 200):
                 rsp = uploadthread.rsp
-                errmsg = ''
+                errmsg = f'Update attempt resulted in response status {uploadthread.rspstatus}'
                 try:
                     rsp = json.loads(rsp)
                     errmsg = (
                         rsp['error'][
                             '@Message.ExtendedInfo'][0]['Message'])
                 except Exception:
-                    raise Exception(uploadthread.rsp)
+                    errmsg = f'Update attempt resulted in response status {uploadthread.rspstatus}: "{repr(rsp)}"'
+                    raise Exception(errmsg)
                 raise Exception(errmsg)
             rsp = json.loads(uploadthread.rsp)
             monitorurl = rsp['@odata.id']
             complete = False
             phase = "apply"
             statetype = 'TaskState'
-            while not complete:
-                pgress = self._do_web_request(monitorurl, cache=False)
+            # sometimes we get an empty pgress when transitioning from the apply phase to
+            # the validating phase; add a retry here so we don't exit the loop in this case
+            retry = 3
+            while not complete and retry > 0:
+                try:
+                    pgress = self._do_web_request(monitorurl, cache=False)
+                except socket.timeout:
+                    pgress = None
                 if not pgress:
-                    break
+                    retry -= 1
+                    time.sleep(3)
+                    continue
+                retry = 3 # reset retry counter
                 for msg in pgress.get('Messages', []):
                     if 'Verify failed' in msg.get('Message', ''):
                         raise Exception(msg['Message'])
@@ -1303,6 +1387,8 @@ class OEMHandler(generic.OEMHandler):
                         time.sleep(3)
                 else:
                     time.sleep(3)
+            if not retry:
+                raise Exception('Falied to monitor update progress due to excessive timeouts')
             if bank == 'backup':
                 return 'complete'
             return 'pending'
@@ -1517,8 +1603,12 @@ class OEMHandler(generic.OEMHandler):
         return 'pending'
 
     def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
-        self.wc.grab_json_response('/api/providers/ffdc',
-                                   {'Generate_FFDC': 1})
+        rsp = self.wc.grab_json_response('/api/providers/ffdc',
+                                         {'Generate_FFDC': 1})
+        if rsp.get('return', 0) == 4:
+            rsp = self.wc.grab_json_response('/api/providers/ffdc',
+                                             {'Generate_FFDC': 1,
+                                              'thermal_log': 0})
         percent = 0
         while percent != 100:
             time.sleep(3)
@@ -1548,7 +1638,7 @@ class OEMHandler(generic.OEMHandler):
             progress({'phase': 'complete'})
         return savefile
 
-    def get_licenses(self):
+    def get_licenses(self, fishclient):
         licdata = self.wc.grab_json_response('/api/providers/imm_fod')
         for lic in licdata.get('items', [{}])[0].get('keys', []):
             if lic['status'] == 0:
@@ -1559,7 +1649,7 @@ class OEMHandler(generic.OEMHandler):
                     'state': 'Missing required license'
                 }
 
-    def delete_license(self, name):
+    def delete_license(self, name, fishclient):
         licdata = self.wc.grab_json_response('/api/providers/imm_fod')
         for lic in licdata.get('items', [{}])[0].get('keys', []):
             if lic.get('feature', None) == name:
@@ -1572,7 +1662,7 @@ class OEMHandler(generic.OEMHandler):
                 )
                 break
 
-    def save_licenses(self, directory):
+    def save_licenses(self, directory, fishclient):
         licdata = self.wc.grab_json_response('/api/providers/imm_fod')
         for lic in licdata.get('items', [{}])[0].get('keys', []):
             licid = ','.join((str(lic['type']), str(lic['id'])))
@@ -1589,7 +1679,7 @@ class OEMHandler(generic.OEMHandler):
                     self._refresh_token()
                 yield savefile
 
-    def apply_license(self, filename, progress=None, data=None):
+    def apply_license(self, filename, fishclient, progress=None, data=None):
         license_errors = {
             310: "License is for a different model of system",
             311: "License is for a different system serial number",
@@ -1613,9 +1703,9 @@ class OEMHandler(generic.OEMHandler):
             if rsp.get('return', 0) in license_errors:
                 raise pygexc.InvalidParameterValue(
                     license_errors[rsp['return']])
-        return self.get_licenses()
+        return self.get_licenses(fishclient)
 
-    def user_delete(self, uid):
+    def user_delete(self, uid, fishclient=None):
         userinfo = self.wc.grab_json_response('/api/dataset/imm_users')
         uidtonamemap = {}
         for user in userinfo.get('items', [{'users': []}])[0].get('users', []):
@@ -1641,6 +1731,11 @@ class OEMHandler(generic.OEMHandler):
         yield "System"
         for key in natural_sort(hwmap):
             yield key
+        for cpuinv in self._get_cpu_inventory():
+            yield cpuinv[0]
+        for meminv in self._get_mem_inventory():
+            yield meminv[0]
+
 
     def get_inventory_of_component(self, compname):
         if compname.lower() == 'system':
@@ -1657,7 +1752,12 @@ class OEMHandler(generic.OEMHandler):
         try:
             return hwmap[compname]
         except KeyError:
-
+            for cpuinv in self._get_cpu_inventory():
+                if cpuinv[0] == compname:
+                    return cpuinv[1]
+            for meminv in self._get_mem_inventory():
+                if meminv[0] == compname:
+                    return meminv[1]
             return None
 
     def get_inventory(self, withids=False):
